'use strict';
const tapahtumaTyypit = {
	lisääAste: 'lisääAste',
	poistaAste: 'poistaAste',
	muutaAste: 'muutaAste',

	lisääLuokka: 'lisääLuokka',
	poistaLuokka: 'poistaLuokka',

	lisääOpettaja: 'lisääOpettaja',
	poistaOpettaja: 'poistaOpettaja',

	lisääTunti: 'lisääTunti',
	poistaTunti: 'poistaTunti',
};

class Tapahtuma {
	constructor(tyyppi, argumentit) {
		this.tyyppi = tyyppi;
		this.argumentit = argumentit;
	}
}

let historia, tulevaisuus;
let luokkaAsteet, opettajat, tunnit;
alustaMalli();

function alustaMalli() {
	historia = [];
	tulevaisuus = [];
	luokkaAsteet = new LuokkaAsteet();
	opettajat = new Opettajat();
	tunnit = new Tunnit();
}

function suorita(tyyppi, ...argumentit) {
	let paluuarvo = undefined;
	switch (tyyppi) {
		case tapahtumaTyypit.lisääAste:
			assertRange('lisääAste argumentit määrä', argumentit.length, 0, 1);
			paluuarvo = luokkaAsteet.lisää(...argumentit)
			break;
		case tapahtumaTyypit.poistaAste:
			assertEq('poistaAste argumentit määrä', argumentit.length, 1);
			let [poistettuAste] = argumentit;
			luokkaAsteet.poista(poistettuAste);
			// Poista poistettujen luokka-asteiden luokat tunneista
			for (let [id, tunti] of tunnit.tunnit) {
				let luokat = [];
				for (let luokka of tunti.luokat) {
					let luokanAste = parseInt(luokka[0]);
					if (luokanAste !== poistettuAste) {
						luokat.push(luokka);
					}
				}
				tunnit.tunnit.get(id).luokat = luokat;
			}
			break;
		case tapahtumaTyypit.muutaAste:
			assertEq('muutaAste argumentit määrä', argumentit.length, 2);
			let [vanhaAste, uusiAste] = argumentit;
			luokkaAsteet.muuta(vanhaAste, uusiAste);
			// Muuta muutetut luokka-asteet tunneissa
			for (let [id, tunti] of tunnit.tunnit) {
				let luokat = [];
				for (let luokka of tunti.luokat) {
					let luokanAste = parseInt(luokka[0]);
					if (luokanAste === vanhaAste) {
						let luokanLuokka = luokka.slice(1);
						luokat.push(`${uusiAste}${luokanLuokka}`);
					} else {
						luokat.push(luokka);
					}
				}
				tunnit.tunnit.get(id).luokat = luokat;
			}
			break;

		case tapahtumaTyypit.lisääLuokka:
			assertEq('lisääLuokka argumentit määrä', argumentit.length, 1);
			luokkaAsteet.asteet[argumentit[0]].lisää();
			break;
		case tapahtumaTyypit.poistaLuokka:
			assertEq('poistaLuokka argumentit määrä', argumentit.length, 1);
			let [aste] = argumentit;
			luokkaAsteet.asteet[aste].poista();
			// Poista luokka jota ei enää ole asteella tunneista
			for (let [id, tunti] of tunnit.tunnit) {
				let luokat = [];
				for (let luokka of tunti.luokat) {
					if (luokkaAsteet.luokat().indexOf(luokka) !== -1) {
						luokat.push(luokka);
					}
				}
				tunnit.tunnit.get(id).luokat = luokat;
			}
			break;

		case tapahtumaTyypit.lisääOpettaja:
			assertEq('lisääOpettaja argumentit määrä', argumentit.length, 2);
			opettajat.lisää(...argumentit);
			break;
		case tapahtumaTyypit.poistaOpettaja:
			assertEq('poistaOpettaja argumentit määrä', argumentit.length, 1);
			let [poistettuOpettaja] = argumentit;
			opettajat.poista(poistettuOpettaja);
			// Poista opettaja joita ei enää ole tunneista
			for (let [id, tunti] of tunnit.tunnit) {
				let opettajaLyhenteet = [];
				for (let opettajaLyhenne of tunti.opettajaLyhenteet) {
					if (opettajaLyhenne !== poistettuOpettaja) {
						opettajaLyhenteet.push(opettajaLyhenne);
					}
				}
				tunnit.tunnit.get(id).opettajaLyhenteet = opettajaLyhenteet;
			}
			break;

		case tapahtumaTyypit.lisääTunti:
			assertEq('lisääTunti argumentit määrä', argumentit.length, 4);
			paluuarvo = tunnit.lisää(...argumentit);
			break;
		case tapahtumaTyypit.poistaTunti:
			assertEq('poistaTunti argumentit määrä', argumentit.length, 1);
			tunnit.poista(...argumentit);
			break;

		default:
			throw new Error(`tuntematon tapahtumatyyppi ${tyyppi}`);
	}
	historia.push(new Tapahtuma(tyyppi, argumentit));
	tulevaisuus = [];
	return paluuarvo;
}

function kumoa() {
	if (historia.length === 0) {
		return;
	}

	// Kumoaminen tapahtuu ottamalla historia uusinta tapahtumaa lukuun
	// ottamatta ja suorittamalla se siihen asti uudestaan tyhjältä mallilta
	let kumottu = historia.pop();
	let uusi_tulevaisuus = tulevaisuus.concat(kumottu);
	let vanha_historia = historia;
	alustaMalli();
	for (let {tyyppi, argumentit} of vanha_historia) {
		suorita(tyyppi, ...argumentit);
	}
	tulevaisuus = uusi_tulevaisuus;
}

function teeUudelleen() {
	if (tulevaisuus.length === 0) {
		return;
	}

	let {tyyppi, argumentit} = tulevaisuus.pop();
	// Tulevaisuus tulee tallentaa, sillä suorita() tuhoaa sen
	let uusi_tulevaisuus = tulevaisuus;
	suorita(tyyppi, ...argumentit);
	tulevaisuus = uusi_tulevaisuus;
}

testi('mallin alustaminen', () => {
	historia = undefined;
	tulevaisuus = undefined;
	luokkaAsteet = undefined;
	opettajat = undefined;
	tunnit = undefined;
	alustaMalli();
	assertNe('historia', historia, undefined);
	assertNe('tulevaisuus', tulevaisuus, undefined);
	assertNe('luokkaAsteet', luokkaAsteet, undefined);
	assertNe('opettajat', opettajat, undefined);
	assertNe('tunnit', tunnit, undefined);
});

testi('tapahtumahistoria', () => {
	alustaMalli();
	suorita(tapahtumaTyypit.lisääAste);
	suorita(tapahtumaTyypit.poistaAste, 1);
	assertEq('historia.length', historia.length, 2);
	assertEq('historia[0].tyyppi', historia[0].tyyppi, tapahtumaTyypit.lisääAste);
	assertEq('historia[0].argumentit.length', historia[0].argumentit.length, 0);
	assertEq('historia[1].tyyppi', historia[1].tyyppi, tapahtumaTyypit.poistaAste);
	assertEq('historia[1].argumentit.length', historia[1].argumentit.length, 1);
	assertEq('historia[1].argumentit[0]', historia[1].argumentit[0], 1);
	assertThrow('poistaAste 1', 'luokka-astetta 1 ei ole olemassa', () => {
		suorita(tapahtumaTyypit.poistaAste, 1);
	});
	assertEq('historia.length poikkeuksen jälkeen', historia.length, 2);
	alustaMalli();
});

testi('kumoamiminen', () => {
	alustaMalli();
	kumoa();
	suorita(tapahtumaTyypit.lisääAste);
	suorita(tapahtumaTyypit.lisääLuokka, 1);
	suorita(tapahtumaTyypit.lisääAste);
	suorita(tapahtumaTyypit.lisääLuokka, 2);
	suorita(tapahtumaTyypit.lisääLuokka, 1);
	assertEq('aluksi 1. aste', luokkaAsteet.asteet[1].luokat(), ['A', 'B', 'C']);
	assertEq('aluksi 2. aste', luokkaAsteet.asteet[2].luokat(), ['A', 'B']);
	kumoa();
	assertEq('kerran 1. aste', luokkaAsteet.asteet[1].luokat(), ['A', 'B']);
	assertEq('kerran 2. aste', luokkaAsteet.asteet[2].luokat(), ['A', 'B']);
	kumoa();
	assertEq('kahdesti 1. aste', luokkaAsteet.asteet[1].luokat(), ['A', 'B']);
	assertEq('kahdesti 2. aste', luokkaAsteet.asteet[2].luokat(), ['A']);
	kumoa();
	assertEq('kolmesti 1. aste', luokkaAsteet.asteet[1].luokat(), ['A', 'B']);
	assertEq('kolmesti 2. aste', luokkaAsteet.asteet[2], undefined);
	kumoa();
	assertEq('neljästi 1. aste', luokkaAsteet.asteet[1].luokat(), ['A']);
	kumoa();
	assertEq('viidesti 1. aste', luokkaAsteet.asteet[1], undefined);
	alustaMalli();
});

testi('uudelleen tekeminen', () => {
	alustaMalli();
	teeUudelleen();
	suorita(tapahtumaTyypit.lisääAste);
	suorita(tapahtumaTyypit.lisääAste);
	suorita(tapahtumaTyypit.lisääLuokka, 1);
	suorita(tapahtumaTyypit.lisääLuokka, 2);
	suorita(tapahtumaTyypit.lisääLuokka, 2);
	kumoa();
	kumoa();
	kumoa();
	assertEq('aluksi 1. aste', luokkaAsteet.asteet[1].luokat(), ['A']);
	assertEq('aluksi 2. aste', luokkaAsteet.asteet[2].luokat(), ['A']);
	teeUudelleen();
	assertEq('kerran 1. aste', luokkaAsteet.asteet[1].luokat(), ['A', 'B']);
	assertEq('kerran 2. aste', luokkaAsteet.asteet[2].luokat(), ['A']);
	teeUudelleen();
	assertEq('kahdesti 1. aste', luokkaAsteet.asteet[1].luokat(), ['A', 'B']);
	assertEq('kahdesti 2. aste', luokkaAsteet.asteet[2].luokat(), ['A', 'B']);
	suorita(tapahtumaTyypit.lisääLuokka, 1);
	teeUudelleen();
	assertEq('kolmesti 1. aste', luokkaAsteet.asteet[1].luokat(), ['A', 'B', 'C']);
	assertEq('kolmesti 2. aste', luokkaAsteet.asteet[2].luokat(), ['A', 'B']);
	alustaMalli();
});

testi('asteiden käsittely', () => {
	alustaMalli();
	assertEq('lisää', suorita(tapahtumaTyypit.lisääAste), 1);
	suorita(tapahtumaTyypit.muutaAste, 1, 2);
	assertEq('muutettua seuraava aste', luokkaAsteet.seuraavaAste(), 3);
	suorita(tapahtumaTyypit.poistaAste, 2);
	assertEq('lopuksi seuraava aste', luokkaAsteet.seuraavaAste(), 1);
	alustaMalli();
});

testi('luokkien käsittely', () => {
	alustaMalli();
	suorita(tapahtumaTyypit.lisääAste);
	assertEq('aluksi', luokkaAsteet.asteet[1].luokat(), ['A']);
	suorita(tapahtumaTyypit.lisääLuokka, 1);
	assertEq('lisättyä', luokkaAsteet.asteet[1].luokat(), ['A', 'B']);
	suorita(tapahtumaTyypit.poistaLuokka, 1);
	assertEq('poistettua', luokkaAsteet.asteet[1].luokat(), ['A']);
	alustaMalli();
});

testi('opettajien käsittely', () => {
	alustaMalli();
	assertEq('aluksi', opettajat.opettajat(), []);
	suorita(tapahtumaTyypit.lisääOpettaja, 'MM', 'Maija Meikäläinen');
	assertEq('lisättyä', opettajat.opettajat(), [['MM', 'Maija Meikäläinen']]);
	suorita(tapahtumaTyypit.poistaOpettaja, 'MM');
	assertEq('poistettua', opettajat.opettajat(), []);
});

testi('tuntien käsittely', () => {
	alustaMalli();
	suorita(tapahtumaTyypit.lisääAste);
	suorita(tapahtumaTyypit.lisääAste, 5);
	suorita(tapahtumaTyypit.lisääLuokka, 5);
	suorita(tapahtumaTyypit.lisääOpettaja, 'KV', 'Kari Virtanen');
	suorita(tapahtumaTyypit.lisääOpettaja, 'AS', 'Aili Savolainen');
	suorita(tapahtumaTyypit.lisääOpettaja, 'MM', 'Maija Meikäläinen');
	assertEq('aluksi', tunnit.järjestyksessä(), []);

	assertEq('historia',
		suorita(tapahtumaTyypit.lisääTunti, 'Historia', '5B', 'KV', 1),
		0
	);
	assertEq('äidinkieli',
		suorita(tapahtumaTyypit.lisääTunti, 'Äidinkieli', '1A', 'AS', 3),
		1
	);
	assertEq('kuvataide',
		suorita(tapahtumaTyypit.lisääTunti, 'Kuvataide', '5A', 'MM', 1),
		2
	);
	assertEq('lisättyä pituus', tunnit.järjestyksessä().length, 3);

	suorita(tapahtumaTyypit.poistaAste, 1);
	assertEq('aste poistettua 0', tunnit.järjestyksessä()[0][1].luokat, ['5B']);
	assertEq('aste poistettua 1', tunnit.järjestyksessä()[1][1].luokat, ['5A']);
	assertEq('aste poistettua 2', tunnit.järjestyksessä()[2][1].luokat, []);

	suorita(tapahtumaTyypit.muutaAste, 5, 6);
	assertEq('aste muutettua 0', tunnit.järjestyksessä()[0][1].luokat, ['6B']);
	assertEq('aste muutettua 1', tunnit.järjestyksessä()[1][1].luokat, ['6A']);
	assertEq('aste muutettua 2', tunnit.järjestyksessä()[2][1].luokat, []);

	suorita(tapahtumaTyypit.poistaLuokka, 6);
	assertEq('luokka poistettua 0', tunnit.järjestyksessä()[0][1].luokat, []);
	assertEq('luokka poistettua 1', tunnit.järjestyksessä()[1][1].luokat, ['6A']);
	assertEq('luokka poistettua 1', tunnit.järjestyksessä()[2][1].luokat, []);

	suorita(tapahtumaTyypit.poistaOpettaja, 'KV');
	assertEq('opettaja poistettua 0', tunnit.järjestyksessä()[0][1].opettajaLyhenteet, []);
	assertEq('opettaja poistettua 1', tunnit.järjestyksessä()[1][1].opettajaLyhenteet, ['MM']);
	assertEq('opettaja poistettua 2', tunnit.järjestyksessä()[2][1].opettajaLyhenteet, ['AS']);

	suorita(tapahtumaTyypit.poistaTunti, 0);
	assertEq('poistettua pituus', tunnit.järjestyksessä().length, 2);

	alustaMalli();
	// TODO: Luokkien ja opettajien listojen päivitys
});
